from epub_conversion import convert_wiki_to_lines
from epub_conversion.wiki_decoder import almost_smart_open
from .utils import line_converter
from .dump_result import DumpResult

def parse_dump(path, max_articles = 1000):
	"""
	Convert a dump to a set of articles with their
	text tokenized and the intrawiki links separated and
	matched with the corresponding sentences.
	Also creates a unique id over the articles for better
	memory footprint (though overall this script is probably
	notoriously bad at its memory management)
	

	Inputs
	------
	        path str : the location of the wiki
	                   dump (bz2 or xml file).
	max_articles int : the number of [valid] articles
	                   to be read before stopping

	
	Output
	------
	result DumpResult : the result of the parse.

	"""
	result = DumpResult()
	wiki = almost_smart_open(path)

	for line, article_name, links in convert_wiki_to_lines(
		wiki,
		max_articles = max_articles,
		line_converter = line_converter):
			result.observe_line(line, article_name, links)

	return result


from .parsed_page import ParsedPage

class ParsedPageParent(ParsedPage):
	"""
	Stores the content of a parsed page. Specialized for
	parent pages (that can have children pages, e.g. leaves
	/ articles)
	
	"""

	def __init__(self, *args):
		super().__init__(*args)
		self.children = set()

	def add_child(self, child):
		"""
		Register a new child for a category
		"""
		self.children.add(child)
		ParsedPage.categories_counter.update([self.name])


"""
Module for Container classes useful for storing metadata
and graph relations within the wiki stored in a wiki dump.

"""

from .parsed_page import ParsedPage
from .parsed_page_child import ParsedPageChild
from .parsed_page_parent import ParsedPageParent

__all__ = ["ParsedPage", "ParsedPageChild", "ParsedPageParent"]


from .parsed_page import ParsedPage
from .parsed_page_parent import ParsedPageParent

class ParsedPageChild(ParsedPage):
	"""
	Stores the content of a parsed page. Specialized for
	child pages (that can have parent pages, e.g. categories)
	
	"""
	def __init__(self, *args):
		super().__init__(*args)
		self.lines = []
		self.parents = set()

	def add_parents(self, parents):
		"""
		Add parents to a parsed page

		Inputs
		------

		parents list : a list of parent ids that correspond to this page

		"""
		for parent, parent_id in parents:
			if ParsedPage.categories.get(parent, None):
				parent_page = ParsedPage.categories[parent]
			else:
				parent_page = ParsedPageParent(parent, parent_id)
				ParsedPage.categories[parent] = parent_page
			
			parent_page.add_child(self.id)

		self.parents.update([link[1] for link in parents])


from collections import Counter

class ParsedPage:
	"""
	Stores the content of a parsed page
	"""

	categories = {}
	categories_counter = Counter()

	def __init__(self, name, id):
		"""
		Add a name and an id to a parsed page
		for referencing.
		"""
		self.id      = id
		self.name    = name


import xml_cleaner.wiki_markup_processing as wiki_processing
double_bracket_no_other_ref = wiki_processing.re.compile("\[\[([^\|\]]+)\]\]")

def get_article_links(sentence):
	"""
	Uses regular expressions to extract
	links from text by looking for double brackets
	and optionally pipe characters to symbolic
	target and anchor text in wiki markup syntax

	Inputs
	------

	sentence str: text to parse

	Output
	------

	list<tuple<str, str>> : list of tuples of target and anchor texts

	"""
	pairs_of_anchor_tags = [(target.strip(), anchor.strip()) for target, anchor in wiki_processing.anchortag_internal_link.findall(sentence)]
	other_links = [(match.strip(), match.strip()) for match in double_bracket_no_other_ref.findall(sentence)]
	return pairs_of_anchor_tags + other_links

def line_converter(lines, article_name):
	"""
	Creates generator that takes the valid text from wiki xml dumps 
	and uses regular expressions to select only the lines
	in the wiki that uses intra wiki links and returns the triplets
	of valid tokenized sentence, article name, and tuples of targets
	and anchor text from intra wiki links.

	Inputs
	------

	lines str: a block of text from a wiki
	article_name str: the name of the current article


	Outputs
	-------

	tuple<str, str, list<tuple<str, str>>> : triplets
	of valid tokenized sentence, article name, and tuples of targets
	and anchor text from intra wiki links.

	"""
	for sentence in wiki_processing.to_raw_text_pairings(lines):
		joined_sentence = " ".join(sentence)
		links = get_article_links(joined_sentence)
		if len(links) > 0:
			yield((wiki_processing.remove_brackets(joined_sentence).split(" "), article_name, links))


from collections import Counter
from .pages import ParsedPageChild

class DumpResult:
	"""
	Stores the result of a dump parse.
	Contains the lines observed with their intra wiki links,
	the counters for each article seen,
	and the targets along with a mapping to a unique integer id.
	"""

	def __init__(self):
		"""
		Stores the result of a dump parse.
		Contains the lines observed with their intra wiki links,
		the counters for each article seen,
		and the targets along with a mapping to a unique integer id.
		"""
		self.stored_lines = {}
		self.target_counters = Counter()
		self.targets = {}
		self.index2target = []

	def observe_line(self, line, article_name, links):
		"""
		Adds a set of line to the result by integrating the
		article name, the links, and converting the links to their
		unique ids as stored previously.

		Inputs
		------

		             line list<str> : a list of strings with double bracket
		                              wiki syntax inter wiki links
		           article_name str : the name of the article these links come
		                              from.
		links list<tuple<str, str>> : a pairing of target and anchor text

		"""

		if self.stored_lines.get(article_name, None) == None:

			# get the id of this article:
			if self.targets.get(article_name):
				article_id = self.targets[article_name]
			else:
				article_id = len(self.targets)
				self.targets[article_name] = article_id
				self.index2target.append(article_name)

			page = ParsedPageChild(article_name, article_id)
			self.stored_lines[article_name] = page
		else:
			page = self.stored_lines[article_name]

		page.lines.append(
			(line, list(self.replace_links_with_index(links)))
		)

		page.add_parents([(link[0], self.targets[link[0]]) for link in links if link[0].startswith("Category")])
		self.target_counters.update((link[0] for link in links))

	def replace_links_with_index(self, links):
		"""
		Takes a set of tuples of targets and anchor text and replaces the
		targets by their integer id from the dump result.

		Inputs
		------
		
		links list<tuple<str, str>> : a pairing of target and anchor text

		"""
		for target, anchor in links:
			if self.targets.get(target):
				yield((self.targets[target], anchor))
			else:
				self.targets[target] = len(self.targets)
				self.index2target.append(target)
				yield((self.targets[target], anchor))


class ParsedPage:
	"""
	Stores the content of a parsed page
	"""

	categories = {}

	def __init__(self, name, id):
		"""
		Add a name and an id to a parsed page
		for referencing.
		"""
		self.id      = id
		self.name    = name
		self.lines   = []
		self.parents = set()

	def add_parents(self, parents):
		"""
		Add parents to a parsed page

		Inputs
		------

		parents list : a list of parent ids that correspond to this page

		"""
		for parent, parent_id in parents:
			if ParsedPageParent.categories.get(parent, None):
				parent_page = ParsedPageParent.categories[parent]
			else:
				parent_page = ParsedPageParent(parent, parent_id)
				ParsedPageParent.categories[parent] = parent_page
			
			parent_page.add_child(self.id)


		self.parents.update(parents)


import os
from setuptools import setup, find_packages

def readfile(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='wikipedia-ner',
    version='0.0.4',
    description='Python package for creating labeled examples from wiki dumps',
    long_description=readfile('README.md'),
    ext_modules=[],
    packages=find_packages(),
    py_modules = ["epub_conversion"],
    author='Jonathan Raiman',
    author_email='jraiman at mit dot edu',
    url='https://github.com/JonathanRaiman/wikipedia_ner',
    download_url='https://github.com/JonathanRaiman/wikipedia_ner',
    keywords='XML, epub, tokenization, NLP',
    license='MIT',
    platforms='any',
    zip_safe=False,
    classifiers=[
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.3',
        'Topic :: Text Processing :: Linguistic',
    ],
    # test_suite="something.test",
    setup_requires = [],
    install_requires=[
        'xml_cleaner',
        'epub_conversion'
    ],
    include_package_data=True,
)


"""
Parsing sub module for converting wiki xml dump pages to labeled examples
"""

from .parse import parse_dump
from .dump_result import DumpResult
from .pages import ParsedPage, ParsedPageChild, ParsedPageParent

__all__ = ["parse_dump", "DumpResult","ParsedPage", "ParsedPageChild", "ParsedPageParent"]